/*
 * LTSP Swap server
 *
 * ltspswapd is based heavily on nbd-search from Pavel Machek.
 * In fact, I copied nbd-server, and then modified it an awful
 * lot, to do the things that we need in ltsp, like automatically
 * create the swap files.
 *
 * It still uses the NBD protocol, and nbd-client is still used
 * on the thin client, to connect with ltspswapd.
 *
 * Copyright 2005 James McQuillan <jam@Ltsp.org>
 * Copyright 1996-1998 Pavel Machek, distribute under GPL
 *  <pavel@atrey.karlin.mff.cuni.cz>
 * Copyright 2001-2004 Wouter Verhelst <wouter@debian.org>
 * Copyright 2002 Anton Altaparmakov <aia21@cam.ac.uk>
 */

/* Includes LFS defines, which defines behaviours of some of the following
 * headers, so must come before those */

#include "config.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>		/* wait */

#include <signal.h>		/* sigaction */
#include <netinet/tcp.h>
#include <netinet/in.h>		/* sockaddr_in, htons, in_addr */

#include <syslog.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <strings.h>
#include <dirent.h>
#include <errno.h>

#define MY_NAME "ltspswapd"

#include "cliserv.h"


#ifdef DODBG
#define DEBUG( a ) printf( a )
#define DEBUG2( a,b ) printf( a,b )
#define DEBUG3( a,b,c ) printf( a,b,c )
#else
#define DEBUG( a )
#define DEBUG2( a,b ) 
#define DEBUG3( a,b,c ) 
#endif


#define BUFSIZE (1024*1024) /**< Size of buffer that can hold requests */

unsigned int timeout = 0; /**< disconnect timeout */

off_t lastpoint = (off_t)-1;	/**< keep track of where we are in the file, to
				  avoid an lseek if possible */

int debug = 0;

unsigned int port = 9210;	/**< Port I'm listening at */

char *exportname;		/**< File I'm exporting */
char swappath[256];		/* directory where swapfiles reside */

#define RECSIZE 10240

char buf[RECSIZE];

unsigned  long swapsize;

char clientname[256] ;

char pidfname[]      = "/var/run/ltspswapd.pid";
char swapsizestr[30] = "64m";   // Default swap size is 64mb

#define MAXCHILDREN 1024

struct {
  in_addr_t  s_addr;
  pid_t      pid;
} children[MAXCHILDREN];

int setup_swapfile(void);

//-----------------------------------------------------------------------------

/**
 * Read data from a file descriptor into a buffer
 *
 * @param f a file descriptor
 * @param buf a buffer
 * @param len the number of bytes to be read
 **/

inline void
readit(int f, void *buf, size_t len) {
  ssize_t res;
  while (len > 0) {
    DEBUG("*");
    if ((res = read(f, buf, len)) <= 0){
      err("Read failed: %m");
    }
    len -= res;
    buf += res;
  }
}

//-----------------------------------------------------------------------------

/**
 * Write data from a buffer into a filedescriptor
 *
 * @param f a file descriptor
 * @param buf a buffer containing data
 * @param len the number of bytes to be written
 **/

inline void
writeit(int f, void *buf, size_t len)
{
  ssize_t res;
  while( len > 0 ) {
    DEBUG("+");
    if( ( res = write(f, buf, len) ) <= 0 ){
      err("Send failed: %m");
    }
    len -= res;
    buf += res;
  }
}

//-----------------------------------------------------------------------------
//
//   If we get a request to register a client that we already have in the
//   table, then it most likely indicates that the client died, and didn't
//   tell us. So, we need to kill the child pid that was handling that 
//   client
//

void
register_client( in_addr_t client_addr, pid_t pid ){
  int i;

  i = 0;
  while( i < MAXCHILDREN ){
    if( children[i].s_addr == client_addr ){
      kill( children[i].pid, 15 );
    }
    i++;
  }

  i = 0;
  while( i < MAXCHILDREN ){
    if( children[i].s_addr == 0 ){
      children[i].s_addr = client_addr;
      children[i].pid = pid;
      break;
    }
    i++;
  }
}

//-----------------------------------------------------------------------------

void
deregister_client( pid_t pid ){
  int i;
  i = 0;
  while( i < MAXCHILDREN ){
    if( children[i].pid == pid ){
      children[i].pid    = 0;
      children[i].s_addr = 0;
    }
    i++;
  }
}

//-----------------------------------------------------------------------------

/**
 * Signal handler for SIGCHLD
 * @param s the signal we're handling (must be SIGCHLD, or something
 * is severely wrong)
 **/

void
sigchld_handler(int s)
{
  int status;
  char buf[80];
  pid_t pid;

  while ( ( pid = waitpid(-1, &status, WNOHANG) ) > 0 ) {
    if(WIFEXITED(status)) {
      memset(buf,'\0', 80);
      snprintf(buf, 79, "%d", WEXITSTATUS(status));
      syslog(LOG_INFO, "Child exited with %s", buf);
      deregister_client( pid );
    }
  }
}

//-----------------------------------------------------------------------------

/**
 * Handle SIGTERM and dispatch it to our children
 * @param s the signal we're handling (must be SIGTERM, or something
 * is severely wrong).
 **/
void
sigterm_handler(int s) {
  exit(0);
}

//-----------------------------------------------------------------------------

/**
 * Seek to a position in a file, unless we're already there.
 * @param handle a filedescriptor
 * @param a position to seek to
 **/

void
maybeseek(int handle, off_t a) {

  if (a < 0 || a > swapsize ) {
    err("Can not happen\n");
  }
  if (lastpoint != a) {
    if (lseek(handle, a, SEEK_SET) < 0) {
      err("Can not seek locally!\n");
    }
    lastpoint = a;
  }
  else{
    DEBUG("S");
  }
}

//-----------------------------------------------------------------------------

/**
 * Write an amount of bytes at a given offset to the right file. This
 * abstracts the write-side of the multiple file option.
 *
 * @param a The offset where the write should start
 * @param buf The buffer to write from
 * @param len The length of buf
 * @return The number of bytes actually written, or -1 in case of an error
 **/
int
expwrite(int fd, off_t a, char *buf, size_t len)
{
  ssize_t res;

  maybeseek(fd, a );
  res = write(fd, buf, len);
  return ( res < 0 || (size_t)res != len );
}

//-----------------------------------------------------------------------------

/**
 * Read an amount of bytes at a given offset from the right file. This
 * abstracts the read-side of the copyonwrite stuff, and calls
 * rawexpread() with the right parameters to do the actual work.
 * @param a The offset where the read should start
 * @param buf A buffer to read into
 * @param len The size of buf
 * @return The number of bytes actually read, or -1 in case of an error
 **/

int
expread( int fd, off_t a, char *buf, size_t len)
{
  ssize_t res;

  maybeseek( fd, a );
  res = read( fd, buf, len );

  return ( res < 0 || (size_t)res != len );
}

//-----------------------------------------------------------------------------

/**
 * Do the initial negotiation.
 *
 * @param net A socket to do the negotiation over
 **/
void 
negotiate(int net) {
  char zeros[300];
  u64 size_host;

  memset( zeros, 0, 290 );

  if ( write( net, INIT_PASSWD, 8 ) < 0){
    err( "Negotiation failed: %m" );
  }
  cliserv_magic = htonll(cliserv_magic);
  if( write(net, &cliserv_magic, sizeof(cliserv_magic) ) < 0 ){
    err("Negotiation failed: %m");
  }
  size_host = htonll( (u64)swapsize  );
  if( write( net, &size_host, 8 ) < 0 ){
    err("Negotiation failed: %m");
  }
  if( write( net, zeros, 128 ) < 0 ){
    err("Negotiation failed: %m");
  }
}

//-----------------------------------------------------------------------------

/**
 * Serve a file to a single client.
 *
 * @todo This beast needs to be split up in many tiny little manageable
 * pieces. Preferably with a chainsaw.
 *
 * @param net A network socket, connected to an nbd client
 * @return never
 **/

int
mainloop(int net)
{
  struct nbd_request request;
  struct nbd_reply reply;
  int fd;

#ifdef DODBG
  int i = 0;
#endif

  fd = setup_swapfile();

  negotiate(net);

  DEBUG("Entering request loop!\n");
  reply.magic = htonl(NBD_REPLY_MAGIC);
  reply.error = 0;

  while ( 1 ) {
    char buf[BUFSIZE];
    size_t len;

#ifdef DODBG
    i++;
    printf("%d: ", i);
#endif

    if (timeout) {
      alarm(timeout);
    }

    readit( net, &request, sizeof(request) );
    request.from = ntohll(request.from);
    request.type = ntohl(request.type);

    /* Disconnect request. 2.6 nbd.h defines NBD_CMD_DISC, which we
     * will use (again) at some point in the future, but many
     * people are still using 2.4-based systems, and they need to
     * be able to compile as well.*/

    if ( request.type == 2 ) {
      err("Disconnect request received.") ;
    }

    len = ntohl(request.len);

    if( request.magic != htonl(NBD_REQUEST_MAGIC) ){
      err("Not enough magic.");
    }

    if ( len > BUFSIZE ){
      err("Request too big!");
    }

#ifdef DODBG
    printf("%s from %Lu (%Lu) len %d, ", request.type ? "WRITE" :
           "READ", (unsigned long long)request.from,
           (unsigned long long)request.from / 512, len);
#endif

    memcpy( reply.handle, request.handle, sizeof(reply.handle) );

    if ( (ssize_t)( ( off_t)request.from + len ) > swapsize  ){
      DEBUG("[RANGE!]");
      reply.error = htonl(-1);
      writeit( net, &reply, sizeof( reply ) );
      reply.error = 0;
      lastpoint = -1;
      continue;
    }

    if ( request.type == 1 ) {	/* WRITE */
      DEBUG("wr: net->buf, ");
      readit(net, buf, len);
      DEBUG("buf->exp, ");
      if ( expwrite( fd, request.from, buf, len) ) {
        DEBUG("Write failed: %m" );
        reply.error = htonl(-1);
        writeit( net, &reply, sizeof( reply ) );
        reply.error = 0;
        lastpoint = -1;
        continue;
      }
      lastpoint += len;

      writeit( net, &reply, sizeof( reply ) );

      DEBUG("OK!\n");
      continue;
    }

    /* READ */

    DEBUG("exp->buf, ");
    if (expread( fd, request.from, buf + sizeof(struct nbd_reply), len)) {
      lastpoint = -1;
      DEBUG("Read failed: %m");
      reply.error = htonl(-1);
      writeit( net, &reply, sizeof( reply ) );
      reply.error = 0;
      lastpoint = -1;
      continue;
    }
    lastpoint += len;

    DEBUG("buf->net, ");
    memcpy(buf, &reply, sizeof(struct nbd_reply));
    writeit(net, buf, len + sizeof(struct nbd_reply));
    DEBUG("OK!\n");
  }
}

//-----------------------------------------------------------------------------

unsigned long 
parse_number(const char *numstr) {
    unsigned long int num;
    char *p;

    num = strtoul(numstr, &p, 10);

    while ( *p ) {
      switch(*p) {
        case 'k':     num *= 1024;
                      ;;
        case 'm':     num *= 1048576;
                      ;;
      }
      p++;
    }
    return( num );
}

//-----------------------------------------------------------------------------

int
setup_swapfile() {

  char      swapfile[300];
  struct    stat statinfo;
  int       create_swapfile = 0;
  long      nCurSize;
  long      nRemain;
  long      nReqSize;
  long      nWritten;
  int       fd;

  if( swappath[strlen(swappath)-1] == '/' ){
    sprintf( swapfile, "%s%s.swap", swappath, clientname );
  }
  else{
    sprintf( swapfile, "%s/%s.swap", swappath, clientname );
  }

  swapsize = parse_number(swapsizestr);

  if( stat( swapfile, &statinfo ) < 0 ){
    if( errno == ENOENT ){
      create_swapfile = 1;
    }
    else{
      fprintf(stderr,"ltspswapd failed: %s\n", strerror(errno));
    }
  }
  else{
    if( statinfo.st_size != swapsize ){
      syslog( LOG_INFO,
              "Removing exising swapfile for %s, it's the wrong size: %ld",
              clientname, statinfo.st_size );
      unlink( swapfile );
      create_swapfile = 1;
    }
  }

  if( create_swapfile ){
    syslog( LOG_INFO, "Creating new swapfile for %s, size: %ld",
                      clientname, swapsize );
    if( ( fd = open(swapfile,O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR)) < 0){
      fprintf( stderr, "Error creating swapfile: %s\n", strerror(errno) );
      return(-1);
    }
    nCurSize = 0L;
    memset(buf,0,RECSIZE);
    while( nCurSize < swapsize ){
      nRemain  = swapsize - nCurSize;
      nReqSize = ( nRemain > RECSIZE ) ? RECSIZE : nRemain;
      nWritten = write( fd, buf, nReqSize );
      if( nWritten != nReqSize ){
        fprintf(stderr,"Error writing to %s: %s\n", swapfile, strerror(errno));
        exit(1);
      }
      nCurSize += nReqSize;
    }
    close( fd );
  }

  if( ( fd = open( swapfile, O_RDWR ) ) < 0 ){
    fprintf( stderr, "Error creating swapfile: %s\n", strerror(errno) );
    return(-1);
  }

  return( fd );
}

//-----------------------------------------------------------------------------

void
wait_for_connection(unsigned int port) {
  struct sockaddr_in addrin;
  struct sigaction sa;
  size_t addrinlen = sizeof(addrin);
  int    net, sock;
  FILE *pidf;

#ifndef sun
  int yes=1;
#else
  char yes='1';
#endif /* sun */

  if( !debug ){
    if(port) {
      if( daemon(0,0) < 0 ) {
        err( "daemon" );
      }
      pidf=fopen( pidfname, "w" );
      if( pidf ){
        fprintf(pidf,"%d", (int)getpid() );
        fclose(pidf);
      }
      else{
        perror("fopen");
        fprintf(stderr, "Not fatal; continuing");
      }
    }
    syslog( LOG_INFO, "Started" );
  }

  /*
   * Clear the children table
   */
  memset( children, 0, sizeof(children) );

  if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0){
    err("socket: %m");
  }

  /*
  ** lose the pesky "Address already in use" error message
  */
  if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR, &yes, sizeof(int)) == -1) {
    err("setsockopt SO_REUSEADDR");
  }

  if (setsockopt(sock,SOL_SOCKET,SO_KEEPALIVE, &yes, sizeof(int)) == -1) {
    err("setsockopt SO_KEEPALIVE");
  }

  DEBUG("Waiting for connections... bind, ");

  addrin.sin_family      = AF_INET;
  addrin.sin_port        = htons(port);
  addrin.sin_addr.s_addr = 0;

  if ( bind( sock, (struct sockaddr *) &addrin, addrinlen ) < 0 ){
    err("bind: %m");
  }

  DEBUG("listen, ");
  if ( listen( sock, 10 ) < 0){
    err("listen: %m");
  }

  DEBUG("accept, ");
  sa.sa_handler = sigchld_handler;
  sigemptyset(&sa.sa_mask);

  sa.sa_flags = SA_RESTART;
  if(sigaction(SIGCHLD, &sa, NULL) == -1){
    err("sigaction: %m");
  }

  sa.sa_handler = sigterm_handler;
  sigemptyset(&sa.sa_mask);

  sa.sa_flags = SA_RESTART;
  if(sigaction(SIGTERM, &sa, NULL) == -1)
    err("sigaction: %m");

  while( 1 ) { /* infinite loop */

    struct sockaddr_in addrin;
    size_t addrinlen = sizeof( addrin );
    int    pid;

    if ((net = accept( sock, (struct sockaddr *) &addrin, &addrinlen ) ) < 0 ){
      err("accept: %m");
    }

    DEBUG("\n");

    if ( getpeername( net, (struct sockaddr *) &addrin, &addrinlen ) < 0 ){
      err("getsockname failed: %m");
    }
    strncpy( clientname, inet_ntoa(addrin.sin_addr), 255 );

    syslog( LOG_INFO, "Serving swap for %s", clientname );

    if( ( pid = fork() ) < 0 ) {
      syslog( LOG_ERR, "Could not fork (%m)" ) ;
      close(net) ;
      continue ;
    }
    if( pid > 0 ) { /* parent */
      close(net);
      register_client( addrin.sin_addr.s_addr, pid );
      continue;
    }
    /* child */
    close(sock);    /* Close the socket, because the child doesn't accept
                         new connections. */

    mainloop(net);
    exit(0);
  }
}

//-----------------------------------------------------------------------------

void
usage() {
  fprintf(stderr,"\nUsage: ltspswapd [-p <portnum>]          default=9210\n");
  fprintf(stderr,"                  -s <path_to_swapfiles>\n");
  fprintf(stderr,"                  -z <swapsize>          default=64mb\n");
  fprintf(stderr,"                  -d                     Enable debugging\n");
  fprintf(stderr,"\n");
  exit(1);
}

//-----------------------------------------------------------------------------

void
DecodeCommandLine(int argc, char * const argv[] ){
  signed int c;
  int found_swappath = 0;

  while( ( c = getopt( argc, argv, "p:s:z:dh")) != -1 ){
    switch( c ){

      case 'p':  port = atoi( optarg );
                 break;

      case 's':  strncpy( swappath, optarg, sizeof(swappath) );
                 found_swappath = 1;
                 break;

      case 'z':  strncpy( swapsizestr, optarg, sizeof(swapsizestr) );
                 break;

      case 'd':  debug = 1;
                 break;

      case 'h':  usage();
                 break;

      default:   usage();
                 break;
    }
  }

  if( ! found_swappath ){
    usage();
  }
}

//-----------------------------------------------------------------------------

/**
 * Main entry point...
 **/

int
main(int argc, char * const argv[] ){
  if(sizeof( struct nbd_request ) != 28 ){
    fprintf(stderr,"Bad size of structure. Alignment problems?\n");
    exit(-1) ;
  }

  logging();

  DecodeCommandLine( argc, argv );

  wait_for_connection(port); /* serve infinitely */

  return 0 ;
}
